import shutil
from pathlib import Path


def copy_dir(source: Path, dest: Path):
    """Delete contents of dest then copy everything from source."""
    # Ensure dest exists
    dest.mkdir(parents=True, exist_ok=True)

    # Remove everything inside destination
    for item in dest.iterdir():
        if item.is_dir():
            shutil.rmtree(item)
        else:
            item.unlink()

    # Copy everything from source to destination
    for item in source.iterdir():
        target = dest / item.name
        if item.is_dir():
            shutil.copytree(item, target, dirs_exist_ok=True)
        else:
            shutil.copy2(item, target)


# Paths
root = Path(__file__).resolve().parent.parent
source_examples = root / "tests/examples"
source_xl = root / "tests/xl"
dest_examples = Path(__file__).resolve().parent / "examples"
dest_xl = Path(__file__).resolve().parent / "xl"

# Sync examples folder
print(f"Syncing {source_examples} → {dest_examples}")
copy_dir(source_examples, dest_examples)

# Sync xl folder
copy_dir(source_xl, dest_xl)



import logging

# use cmap (which has Bang Wong built in)
from palettable.tableau import (
    BlueRed_6,
    BlueRed_12,
    GreenOrange_6,
    GreenOrange_12,
    PurpleGray_6,
    PurpleGray_12,
    Tableau_10,
    Tableau_20,
)
from pydantic_extra_types.color import Color

BANG_WONG_COLORS = dict(
    black=Color("RGB(0, 0, 0)"),
    orange=Color("RGB(230, 159, 0)"),
    sky_blue=Color("RGB(86, 180, 233)"),
    bluish_green=Color("RGB(0, 158, 115)"),
    yellow=Color("RGB(240, 228, 66)"),
    blue=Color("RGB(0, 114, 178)"),
    vermillion=Color("RGB(213, 94, 0)"),
    reddish_purple=Color("RGB(204, 121, 167)"),
)


class BangWong_7:
    def __init__(self):
        self.colors = BANG_WONG_COLORS

    @property
    def hex_colors(self):
        return [v.as_hex() for k, v in self.colors.items() if k != "black"]


class BangWong_8:
    def __init__(self):
        self.colors = BANG_WONG_COLORS

    @property
    def hex_colors(self):
        return [v.as_hex() for k, v in self.colors.items()]


XLSXDATAGRID_STANDARD_PALLETTES = {
    ("bangwong", (7, 8)): (BangWong_7(), BangWong_8()),
    ("BlueRed", (6, 12)): (BlueRed_6, BlueRed_12),
    ("PurpleGray", (6, 12)): (PurpleGray_6, PurpleGray_12),
    ("GreenOrange", (6, 12)): (GreenOrange_6, GreenOrange_12),
    ("Tableau", (10, 20)): (Tableau_10, Tableau_20),
}


def get_color_pallette(
    length, palettes_in_use, palettes=XLSXDATAGRID_STANDARD_PALLETTES
):
    _max = list(XLSXDATAGRID_STANDARD_PALLETTES.keys())[-1][1][1]
    _max_pallete = list(palettes.values())[-1][1]
    for k, v in palettes.items():
        if k[0] not in palettes_in_use:
            if length < k[1][0]:
                palettes_in_use += [k[0]]
                return v[0].hex_colors[0:length]
            elif k[1][0] < length <= k[1][1]:
                palettes_in_use += [k[0]]
                return v[1].hex_colors[0:length]
            elif k[1][1] < length <= _max:
                pass
            elif length > _max:
                logging.warning(f"don't have a colour pallette of length: {length}")
                extra = length - _max
                if extra > _max:
                    raise ValueError(
                        f"error selecting colour pallette of length = {length}"
                    )
                else:
                    return _max_pallete.hex_colors + _max_pallete.hex_colors[0:extra]

            else:
                raise ValueError(
                    f"error selecting colour pallette of length = {length}"
                )


def color_variant(hex_color, brightness_offset=1):
    """takes a color like #87c95f and produces a lighter or darker variant

    Reference:
        https://chase-seibert.github.io/blog/2011/07/29/python-calculate-lighterdarker-rgb-colors.html
    """
    if len(hex_color) != 7:
        if len(hex_color) != 4:
            raise Exception(
                "Passed %s into color_variant(), needs to be in #87c95f format."
                % hex_color
            )
        else:
            hex_color = hex_color + hex_color[1:]
    rgb_hex = [hex_color[x : x + 2] for x in [1, 3, 5]]
    new_rgb_int = [int(hex_value, 16) + brightness_offset for hex_value in rgb_hex]
    new_rgb_int = [
        min([255, max([0, i])]) for i in new_rgb_int
    ]  # make sure new values are between 0 and 255
    # hex() produces "0x88", we want just "88"
    return "#" + "".join([hex(i)[2:] for i in new_rgb_int])



import typing as ty
from datetime import date, datetime, time, timedelta
from enum import StrEnum

import jsonref
import pandas as pd
import pytest
import requests
import xlsxwriter as xw
from dirty_equals import IsInstance
from frictionless import Package, Resource
from jsonref import replace_refs
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    NaiveDatetime,
    # NaiveDate,
    RootModel,
    StringConstraints,
    computed_field,
)
from typing_extensions import Annotated

import xlsxdatagrid.xlsxdatagrid as xdg
from xlsxdatagrid.xlsxdatagrid import (
    DataGridSchema,
    FieldSchema,
    XlTableWriter,
    coerce_schema,
    # write_table,
    convert_dict_arrays_to_list_records,
    convert_list_records_to_dict_arrays,
    convert_records_to_datagrid_schema,
    wb_from_dataframe,
    wb_from_dataframes,
    xdg_from_pydantic_object,
    xdg_from_pydantic_objects,
)

from . import constants as c


class MyColor(StrEnum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class Example(BaseModel):
    a_int: int = Field(1, json_schema_extra=dict(section="numeric"))
    a_constrainedint: Annotated[int, Field(ge=0, le=10)] = Field(
        3, json_schema_extra=dict(section="numeric")
    )
    b_float: float = Field(1.5, json_schema_extra=dict(section="numeric"))
    c_str: str = Field("string", json_schema_extra=dict(section="unicode"))
    c_constrainedstr: Annotated[
        str,
        StringConstraints(
            max_length=10,
        ),
    ] = Field("string", json_schema_extra=dict(section="unicode"))
    d_enum: MyColor = Field(json_schema_extra=dict(section="unicode"))
    e_bool: bool = Field(True, json_schema_extra=dict(section="boolean"))
    f_date: date = Field(date.today(), json_schema_extra=dict(section="datetime"))
    g_datetime: NaiveDatetime = Field(
        datetime.now(), json_schema_extra=dict(section="datetime")
    )
    h_time: time = Field(
        datetime.now().time(), json_schema_extra=dict(section="datetime")
    )
    i_duration: timedelta = Field(
        timedelta(days=0, hours=2, minutes=33, seconds=3),
        json_schema_extra=dict(section="datetime"),
    )

    @computed_field(
        description="calc value",
        json_schema_extra=dict(formula="a_int * b_float", section="numeric"),
    )
    def b_calcfloat(self) -> float:
        return self.a_int * self.b_float


class Example1(BaseModel):
    a_int: int = Field(1, json_schema_extra=dict(section="numeric"))
    a_constrainedint: Annotated[int, Field(ge=0, le=10)] = Field(
        3, json_schema_extra=dict(section="numeric")
    )
    b_float: float = Field(1.5, json_schema_extra=dict(section="numeric"))
    c_str: str = Field("string", json_schema_extra=dict(section="unicode"))
    c_constrainedstr: Annotated[
        str,
        StringConstraints(
            max_length=10,
        ),
    ] = Field("string", json_schema_extra=dict(section="unicode"))
    d_enum: MyColor = Field(json_schema_extra=dict(section="unicode"))


class ExampleArray(RootModel):
    model_config = ConfigDict(
        json_schema_extra=dict(
            datagrid_index_name=("section", "title", "name"), is_transposed=False
        )
    )
    root: list[Example]


class ExampleArray1(RootModel):
    model_config = ConfigDict(
        json_schema_extra=dict(
            datagrid_index_name=("section", "title", "name"), is_transposed=False
        )
    )
    root: list[Example1]


class ExampleArrayTransposed(ExampleArray):
    model_config = ConfigDict(
        json_schema_extra=dict(
            datagrid_index_name=("section", "title", "name"), is_transposed=True
        )
    )


TEST_ARRAY_SCHEMA = {
    "$defs": {
        "MyColor": {
            "enum": ["red", "green", "blue"],
            "title": "MyColor",
            "type": "string",
        },
        "Example": {
            "properties": {
                "a_constrainedint": {
                    "default": 3,
                    "maximum": 10,
                    "minimum": 0,
                    "section": "numeric",
                    "title": "A " "Constrainedint",
                    "type": "integer",
                },
                "a_int": {
                    "default": 1,
                    "section": "numeric",
                    "title": "A Int",
                    "type": "integer",
                },
                "b_calcfloat": {
                    "description": "calc value",
                    "formula": "a_int * b_float",
                    "readOnly": True,
                    "section": "numeric",
                    "title": "B Calcfloat",
                    "type": "number",
                },
                "b_float": {
                    "default": 1.5,
                    "section": "numeric",
                    "title": "B Float",
                    "type": "number",
                },
                "c_constrainedstr": {
                    "default": "string",
                    "maxLength": 10,
                    "section": "unicode",
                    "title": "C " "Constrainedstr",
                    "type": "string",
                },
                "c_str": {
                    "default": "string",
                    "section": "unicode",
                    "title": "C Str",
                    "type": "string",
                },
                "d_enum": {
                    "allOf": [{"$ref": "#/$defs/MyColor"}],
                    "section": "unicode",
                },
                "e_bool": {
                    "default": True,
                    "section": "boolean",
                    "title": "E Bool",
                    "type": "boolean",
                },
                "f_date": {
                    "default": "2024-06-06",
                    "format": "date",
                    "section": "datetime",
                    "title": "F Date",
                    "type": "string",
                },
                "g_datetime": {
                    "default": "2024-06-06T10:42:54.822063",
                    "format": "date-time",
                    "section": "datetime",
                    "title": "G Datetime",
                    "type": "string",
                },
                "h_time": {
                    "default": "10:42:54.822257",
                    "format": "time",
                    "section": "datetime",
                    "title": "H Time",
                    "type": "string",
                },
                "i_duration": {
                    "default": "PT2H33M3S",
                    "format": "duration",
                    "section": "datetime",
                    "title": "I Duration",
                    "type": "string",
                },
            },
            "required": ["d_enum", "b_calcfloat"],
            "title": "Example",
            "type": "object",
        },
    },
    "datagrid_index_name": ("section", "title", "name"),
    "is_transposed": False,
    "items": {"$ref": "#/$defs/Example"},
    "title": "ExampleArrayTransposed",
    "type": "array",
}

TEST_ARRAY_SCHEMA_TRANSPOSED = {
    k: (lambda k, v: True if k == "is_transposed" else v)(k, v)
    for k, v in TEST_ARRAY_SCHEMA.items()
}

ARRAY_DATA = {
    "a_int": [1, 2, 3],
    "a_constrainedint": [3, 3, 3],
    "b_float": [1.5, 2.5, 3.5],
    "c_str": ["string", "asdf", "bluey"],
    "c_constrainedstr": ["string", "string", "string"],
    "d_enum": ["green", "green", "blue"],
    "e_bool": [True, True, False],
    "f_date": ["2024-06-06", "2024-06-06", "2024-06-06"],
    "g_datetime": [
        "2024-06-06T10:08:52.078770",
        "2024-06-06T10:08:52.078770",
        "2024-06-06T10:08:52.078770",
    ],
    "h_time": ["10:08:52.078959", "10:08:52.078959", "10:08:52.078959"],
    "i_duration": ["PT2H33M3S", "PT2H33M3S", "PT2H33M3S"],
    "b_calcfloat": [1.5, 5.0, 10.5],
}

ARRAY_DATA1 = {k: v * 2 for k, v in ARRAY_DATA.items() if k in Example1.model_fields}


def get_test_array(is_transposed=False):
    t1, t2, t3 = (
        Example(d_enum=MyColor.GREEN),
        Example(a_int=2, b_float=2.5, c_str="asdf", d_enum=MyColor.GREEN),
        Example(a_int=3, b_float=3.5, c_str="bluey", d_enum=MyColor.BLUE, e_bool=False),
    )

    if is_transposed:
        return ExampleArrayTransposed([t1, t2, t3])
    else:
        return ExampleArray([t1, t2, t3])


def get_pydantic_test_inputs(is_transposed=False):
    if is_transposed:
        return c.PATH_XL_TRANSPOSED, get_test_array(is_transposed)
    else:
        return c.PATH_XL, get_test_array(is_transposed)


@pytest.fixture(autouse=True)
def ensure_xl_dir():
    c.FDIR.mkdir(parents=True, exist_ok=True)


@pytest.fixture(params=[True, False])
def write_table_test(request):
    fpth, pyd_obj = get_pydantic_test_inputs(is_transposed=request.param)
    fpth.unlink(missing_ok=True)
    fpth, xl_tbl = xdg_from_pydantic_object(pyd_obj, fpth)
    return fpth, xl_tbl


def test_pydantic_object_write_table(write_table_test):
    fpth, xl_tbl = write_table_test
    assert fpth.is_file()


def test_pydantic_objects_write_tables():
    fpth, pyd_obj = get_pydantic_test_inputs(is_transposed=False)
    fpth = c.PATH_XL_MANY_SHEETS
    fpth.unlink(missing_ok=True)
    pyd_obj = ExampleArray(convert_dict_arrays_to_list_records(ARRAY_DATA))
    pyd_obj1 = ExampleArray1(convert_dict_arrays_to_list_records(ARRAY_DATA1))

    fpth = xdg_from_pydantic_objects([pyd_obj, pyd_obj1], fpth)
    assert fpth.is_file()


def get_schema_test_inputs(is_transposed=False):
    if is_transposed:
        return (
            c.PATH_XL_FROM_SCHEMA_TRANSPOSED,
            TEST_ARRAY_SCHEMA_TRANSPOSED,
            ARRAY_DATA,
        )
    else:
        return c.PATH_XL_FROM_SCHEMA, TEST_ARRAY_SCHEMA, ARRAY_DATA


@pytest.mark.parametrize("is_transposed", [True, False])
def test_schema_and_data_write_table(is_transposed):
    from xlsxdatagrid.xlsxdatagrid import (  #  write_table,
        DataGridData,
        DataGridSchema,
        get_xlgrid,
        write_grid,
    )

    fpth, schema, data = get_schema_test_inputs(is_transposed=is_transposed)

    fpth.unlink(missing_ok=True)
    gridschema = DataGridSchema(**convert_records_to_datagrid_schema(schema))
    griddata = DataGridData(root=data)
    xlgrid = get_xlgrid(gridschema, griddata)

    # xl_tbl = XlTableWriter(gridschema=gridschema, data=data)

    workbook = xw.Workbook(str(fpth))
    write_grid(workbook, xlgrid, gridschema, griddata)
    workbook.close()
    assert fpth.is_file()


@pytest.mark.skip(reason="needs the MXF api so skip for CI")
class TestDigitalSchedulesApi:
    def test_schema_and_data_from_digital_schedules_api():
        from xlsxdatagrid.xlsxdatagrid import write_table

        fpth = c.PATH_XL_FROM_API
        response = requests.get(
            "https://aectemplater-dev.maxfordham.com/type_specs/project_revision/1/object/602/grid?override_units=true"
        )
        assert (
            response.status_code == 200
        ), f"API request failed with status code {response.status_code}"

        fpth.unlink(missing_ok=True)
        data = jsonref.replace_refs(response.json(), merge_props=True)
        data["data"] = data["data"] + data["data"]
        data_array = convert_list_records_to_dict_arrays(data["data"])

        gridschema = convert_records_to_datagrid_schema(data["$schema"])
        gridschema["datagrid_index_name"] = ("section", "unit", "name")
        gridschema["is_transposed"] = True

        xl_tbl = XlTableWriter(gridschema=gridschema, data=data_array)
        workbook = xw.Workbook(str(fpth))
        write_table(workbook, xl_tbl)
        workbook.close()
        assert fpth.is_file()


def test_IsInstance():
    class Foo(BaseModel):
        a: str = "a"

    class FooArray(RootModel):
        root: list[Foo] = [Foo()]

    assert Foo() == IsInstance(Foo)
    assert Foo() == IsInstance(BaseModel)
    assert Foo() == IsInstance(Foo, only_direct_instance=True)
    assert Foo() != IsInstance(BaseModel, only_direct_instance=True)
    assert isinstance(Foo(), BaseModel)
    assert issubclass(Foo, BaseModel)
    assert not FooArray() == IsInstance(Foo)


def test_enum_field_schema():
    class Example(BaseModel):
        m_color: Annotated[MyColor, Field(json_schema_extra={"section": "unicode"})]

    item = replace_refs(Example.model_json_schema(), merge_props=True)
    name = "m_color"
    fschema = item.get("properties").get(name) | {"name": name}

    f = FieldSchema(**fschema)
    assert f == IsInstance(FieldSchema)


def test_coerce_schema():
    class Foo(BaseModel):
        a: str = "a"

    class FooArray(RootModel):
        root: list[Foo] = [Foo()]

    schema = {
        "$defs": {
            "Foo": {
                "properties": {"a": {"default": "a", "title": "A", "type": "string"}},
                "title": "Foo",
                "type": "object",
            }
        },
        "default": [{"a": "a"}],
        "items": {"$ref": "#/$defs/Foo"},
        "title": "FooArray",
        "type": "array",
    }

    assert coerce_schema(FooArray) == IsInstance(
        DataGridSchema, only_direct_instance=True
    )
    assert coerce_schema(FooArray()) == IsInstance(
        DataGridSchema, only_direct_instance=True
    )
    assert coerce_schema(schema) == IsInstance(
        DataGridSchema, only_direct_instance=True
    )
    assert coerce_schema(
        DataGridSchema(**convert_records_to_datagrid_schema(schema))
    ) == IsInstance(DataGridSchema, only_direct_instance=True)

    assert coerce_schema(ExampleArrayTransposed) == IsInstance(DataGridSchema)

    assert coerce_schema(ExampleArray) == IsInstance(DataGridSchema)


def test_coerce_data():
    from xlsxdatagrid.xlsxdatagrid import DataGridData, coerce_data

    class Row(BaseModel):
        x: int
        y: str

    class Grid(RootModel):
        root: list[Row]

    data1 = {"x": [1, 2, 3], "y": list("abc")}
    data2 = [{"x": x, "y": y} for x, y in zip(data1["x"], data1["y"])]
    data3 = pd.DataFrame(data2)
    data4 = Grid.model_validate(data2)
    data5 = DataGridData(root=data1)

    for x in [data1, data2, data3, data4, data5]:
        check = coerce_data(x)
        assert isinstance(check, DataGridData)
        assert check == data5
    print("done")


def test_datapackage():
    class Foo(BaseModel):
        a: str = "a"
        b: int = 2
        c: float = 2.3

    class FooArray(RootModel):
        root: list[Foo] = [Foo()]

    schema = coerce_schema(FooArray).model_dump(exclude_none=True, exclude="type")
    data = [{k: n * v for k, v in Foo().model_dump().items()} for n in range(1, 4)]

    Resource(data=data, schema=schema)
    Package(
        resources=[Resource(data=pd.DataFrame(data)) for d in data]
    )  # from arguments
    print("done")


def x_squared():
    x = range(-5, 6)
    y = [_**2 for _ in x]
    return pd.DataFrame({"x": x, "y": y})


def x_cubed():
    x = range(-5, 6)
    y = [_**3 for _ in x]
    return pd.DataFrame({"x": x, "y": y})


def test_wb_from_dataframe():
    fpth = c.PATH_FROM_DF_WITH_CHART
    fpth.unlink(missing_ok=True)
    df = x_squared()
    workbook, xl_tbl, worksheet = wb_from_dataframe(df, fpth)

    # TODO: support xl chart from vega-lite spec?
    #       https://frictionlessdata.io/blog/2017/03/31/data-package-views-proposal/#graph-spec

    chart = workbook.add_chart({"type": "line"})

    chart.add_series(
        {
            "name": "x squared",
            "categories": [worksheet.name] + list(xl_tbl.rng_arrays["x"]),
            "values": [worksheet.name] + list(xl_tbl.rng_arrays["y"]),
        }
    )

    # Add a chart title and some axis labels.
    chart.set_title({"name": "x squared"})
    chart.set_x_axis({"name": "x"})
    chart.set_y_axis({"name": "y"})

    # Set an Excel chart style.
    chart.set_style(11)

    # Add the chart to the chartsheet.
    depth = xl_tbl.gridschema.header_depth + 4
    width = len(xl_tbl.gridschema.fields) + 2
    worksheet.insert_chart(depth, width, chart)
    workbook.close()

    assert fpth.is_file()


def test_wb_from_dataframes():
    fpth = c.PATH_FROM_DF_WITH_MANY_CHARTS
    fpth.unlink(missing_ok=True)
    df_labels = pd.DataFrame(
        {"a": [1, 2, 3, None], "b": list("abcd"), "c": [None, "e", "f", pd.NA]}
    )
    # df_labels = df_labels.fillna("")
    workbook, worksheets, xl_tbls = wb_from_dataframes(
        [x_squared(), x_cubed(), df_labels], fpth
    )
    workbook.close()
    assert fpth.is_file()


@pytest.fixture
def from_json_with_null():
    fpth = c.PATH_FROM_JSON
    fpth.unlink(missing_ok=True)

    class ExampleItem(BaseModel):
        a: ty.Optional[int]
        b: str
        c: ty.Optional[str]

    class ExampleGrid(RootModel):
        root: list[ExampleItem]

    data = [
        {"a": 1, "b": "a", "c": None},
        {"a": 2, "b": "b", "c": "e"},
        {"a": 3, "b": "c", "c": "f"},
        {"a": None, "b": "d", "c": None},
    ]
    xdg.xdg_from_json(data, schema=ExampleGrid, fpth=fpth)
    return fpth, data, ExampleGrid


def test_from_json(from_json_with_null):
    fpth, data, ExampleGrid = from_json_with_null
    assert fpth.is_file()


@pytest.mark.parametrize("is_transposed", [True, False])
def test_from_json_empty_data(is_transposed):
    class ExampleItem(BaseModel):
        a: ty.Optional[int]
        b: str
        c: ty.Optional[str]

    class ExampleGrid(RootModel):
        root: list[ExampleItem]

    fpth = (
        c.PATH_FROM_JSON_EMPTY if is_transposed else c.PATH_FROM_JSON_EMPTY_TRANSPOSED
    )

    data = [dict(a=2, b="b", c=None)]
    xdg.xdg_from_json(data, schema=ExampleGrid, fpth=fpth, is_transposed=is_transposed)
    assert fpth.is_file()


def test_get_xlgrid():
    class ExampleItem(BaseModel):
        a: ty.Optional[int]
        b: str
        c: ty.Optional[str]

    class ExampleGrid(RootModel):
        root: list[ExampleItem]

    data = [dict(a=2, b="b", c=None)]
    from xlsxdatagrid.xlsxdatagrid import XlGrid, coerce_data, coerce_schema, get_xlgrid

    gridschema = coerce_schema(ExampleGrid)
    griddata = coerce_data(data)
    xlgrid = get_xlgrid(gridschema, griddata)

    assert isinstance(xlgrid, XlGrid)
    print("done")


def test_write_grid():
    class ExampleItem(BaseModel):
        a: ty.Optional[int]
        b: str
        c: ty.Optional[str]

    class ExampleGrid(RootModel):
        root: list[ExampleItem]

    data = [dict(a=2, b="b", c=None)] * 3
    from xlsxdatagrid.xlsxdatagrid import (
        XlGrid,
        coerce_data,
        coerce_schema,
        get_xlgrid,
        write_grid,
    )

    gridschema = coerce_schema(ExampleGrid)
    griddata = coerce_data(data)
    xlgrid = get_xlgrid(gridschema, griddata)
    c.PATH_WRITE_GRID.unlink(missing_ok=True)
    workbook = xw.Workbook(str(c.PATH_WRITE_GRID))
    worksheet = write_grid(
        workbook, xlgrid=xlgrid, gridschema=gridschema, data=griddata
    )

    assert isinstance(worksheet, xw.worksheet.Worksheet)
    workbook.close()
    assert c.PATH_WRITE_GRID.is_file()



import pytest
from pytest_examples import CodeExample, EvalExample, find_examples


@pytest.mark.parametrize("example", find_examples("tests/examples"), ids=str)
def test_pytest_example(example: CodeExample, eval_example: EvalExample):
    if eval_example.update_examples:
        print("Running example from:", example.path)
        eval_example.format(example)
        eval_example.run_print_update(example)
    else:
        eval_example.lint(example)
        print("Running example from:", example.path)
        eval_example.run_print_check(example)



import datetime
import json
import typing as ty

import pytest
from pydantic import BaseModel

from xlsxdatagrid.read import (
    pydantic_model_from_json_schema,
    read_csv_string,
    read_csv_string_with_metadata,
    read_excel,
    read_excel_from_metadata,
)
from xlsxdatagrid.xlsxdatagrid import DataGridMetaData

from .constants import PATH_JSONSCHEMA_RAW
from .csv_model import DataTypesArrayTransposed
from .test_xlsxdatagrid import (
    ExampleArray,
    ExampleArrayTransposed,
    from_json_with_null,  # req. fixture  # noqa: F401
    write_table_test,  # req. fixture  # noqa: F401
)

schemas = [ExampleArray.model_json_schema(), ExampleArrayTransposed.model_json_schema()]
schemas = {s["title"]: s for s in schemas}


def _as_delimited(text: str, delimiter: str) -> str:
    return text if delimiter == "\t" else text.replace("\t", delimiter)


# def pydantic_model_from_json_schema(json_schema: str) -> ty.Type[BaseModel]:
#     with TemporaryDirectory() as temporary_directory_name:
#         temporary_directory = Path(temporary_directory_name)
#         file_path = "model.py"
#         module_name = file_path.split(".")[0]
#         output = Path(temporary_directory / file_path)
#         generate(
#             json_schema,
#             input_file_type=InputFileType.JsonSchema,
#             input_filename="example.json",
#             output=output,
#             output_model_type=DataModelType.PydanticV2BaseModel,
#         )
#         spec = importlib.util.spec_from_file_location(module_name, output)
#         module = importlib.util.module_from_spec(spec)
#         sys.modules[module_name] = module
#         spec.loader.exec_module(module)
#     return module.Model


def get_schema(name, schemas=schemas):
    return schemas.get(name)


def get_datamodel(metadata: DataGridMetaData) -> dict:
    return schemas.get(metadata.name)


TEST_JSON_SCHEMA: str = """{
    "type": "object",
    "properties": {
        "number": {"type": "number"},
        "street_name": {"type": "string"},
        "street_type": {"type": "string",
                        "enum": ["Street", "Avenue", "Boulevard"]
                        }
    }
}"""

TEST_JSON_SCHEMA1: str = """{
    "properties": {
        "a_int": {
            "default": 1,
            "section": "numeric",
            "title": "A Int",
            "type": "integer"
        },
        "Abbreviation": {
            "enum": [
                "yellow",
                "red",
                "violet"
            ],
            "type": "string",
            "default": "yellow",
            "description": "pick colour"
        }
    },
    "required": [
        "Abbreviation"
    ],
    "title": "Test",
    "type": "object"
}"""

#


def test_load_model_from_json_schema():
    pydantic_model = pydantic_model_from_json_schema(json.loads(TEST_JSON_SCHEMA))
    assert issubclass(pydantic_model, BaseModel)
    assert isinstance(pydantic_model(), BaseModel)


def test_load_model_from_json_schema_issue2091():
    # TODO: remove this once resolved: https://github.com/koxudaxi/datamodel-code-generator/issues/2091
    pydantic_model = pydantic_model_from_json_schema(json.loads(TEST_JSON_SCHEMA1))
    assert issubclass(pydantic_model, BaseModel)
    assert isinstance(pydantic_model(Abbreviation="yellow"), BaseModel)


def test_get_datamodel():
    metadata = DataGridMetaData(name="ExampleArray", title="Example Array")
    jsonschema = get_datamodel(metadata)
    assert jsonschema["title"] == "ExampleArray"


def test_read_excel_from_metadata(write_table_test):  # noqa: F811
    path, xl_tbl = write_table_test
    obj, metadata = read_excel_from_metadata(path, get_datamodel=get_datamodel)
    assert isinstance(obj, list)
    assert len(obj) == 3
    print("done")


def test_read_excel(write_table_test):  # noqa: F811
    path, xl_tbl = write_table_test
    is_transposed = xl_tbl.gridschema.is_transposed
    model = DataTypesArrayTransposed
    data, errors = read_excel(
        path,
        is_transposed=is_transposed,
        header_depth=xl_tbl.gridschema.header_depth,
        model=model,
    )
    assert isinstance(data, list)
    assert len(data) == 3


@pytest.mark.parametrize("delimiter", ["\t", ","], ids=["tsv", "csv"])
def test_read_csv_string(delimiter):
    string = """#Title=ExampleArray - HeaderDepth=3 - IsTransposed=False - DateTime=2025-10-29 11:07:21.014106 - DatamodelUrl=None											
numeric	numeric	numeric	unicode	unicode	unicode	boolean	datetime	datetime	datetime	datetime	numeric
A Int	A Constrainedint	B Float	C Str	C Constrainedstr	MyColor	E Bool	F Date	G Datetime	H Time	I Duration	B Calcfloat
a_int	a_constrainedint	b_float	c_str	c_constrainedstr	d_enum	e_bool	f_date	g_datetime	h_time	i_duration	b_calcfloat
1	3	1.5	string	string	green	TRUE	2025-10-29	2025-10-29T11:07:19+00:00	11:07:19+00:00	PT2H33M03S	1.5
2	3	2.5	asdf	string	green	TRUE	2025-10-29	2025-10-29T11:07:19+00:00	11:07:19+00:00	PT2H33M03S	5
3	3	3.5	bluey	string	blue	FALSE	2025-10-29	2025-10-29T11:07:19+00:00	11:07:19+00:00	PT2H33M03S	10.5
"""

    model = DataTypesArrayTransposed
    data_string = _as_delimited(string, delimiter)
    data, errors = read_csv_string(
        data_string, False, header_depth=3, model=model, delimiter=delimiter
    )

    assert isinstance(data, list)
    assert len(data) == 3
    print("done")


@pytest.mark.parametrize("delimiter", ["\t", ","], ids=["tsv", "csv"])
def test_read_csv_string_transposed(delimiter):
    string = """type	title	name	row1	row2	row3
numeric	A Int	a_int	1	2	3
numeric	A Constrainedint	a_constrainedint	3	3	3
numeric	B Float	b_float	1.5	2.5	3.5
unicode	C Str	c_str	string	asdf	bluey
unicode	C Constrainedstr	c_constrainedstr	string	string	string
unicode	MyColor	d_enum	green	green	blue
boolean	E Bool	e_bool	TRUE	TRUE	FALSE
datetime	F Date	f_date	2025-10-22	2025-10-22	2025-10-22
datetime	G Datetime	g_datetime	2025-10-22T13:36:16+00:00	2025-10-22T13:36:16+00:00	2025-10-22T13:36:16+00:00
datetime	H Time	h_time	13:36:16+00:00	13:36:16+00:00	13:36:16+00:00
datetime	I Duration	i_duration	PT2H33M3S	PT2H33M3S	PT2H33M3S
numeric	B Calcfloat	b_calcfloat	1.5	5	10.5
"""

    model = DataTypesArrayTransposed
    data_string = _as_delimited(string, delimiter)
    data, errors = read_csv_string(
        data_string, True, header_depth=3, model=model, delimiter=delimiter
    )

    assert isinstance(data, list)
    assert len(data) == 3
    print("done")


@pytest.mark.parametrize("delimiter", ["\t", ","], ids=["tsv", "csv"])
def test_read_csv_string_with_metadata(delimiter):
    string = """#Title=ExampleArray - HeaderDepth=3 - IsTransposed=False - DateTime=2025-10-22 15:15:55.981465 - DatamodelUrl=None
    #some hash string comment that should be ignored												
section	numeric	numeric	numeric	unicode	unicode	unicode	boolean	datetime	datetime	datetime	datetime	numeric
title	A Int	A Constrainedint	B Float	C Str	C Constrainedstr	MyColor	E Bool	F Date	G Datetime	H Time	I Duration	B Calcfloat
name	a_int	a_constrainedint	b_float	c_str	c_constrainedstr	d_enum	e_bool	f_date	g_datetime	h_time	i_duration	b_calcfloat
	1	3	1.5	string	string	green	TRUE	2025-10-22	2025-10-22T15:15:56+00:00	15:15:56+00:00	PT2H33M03S	1.5
	2	3	2.5	asdf	string	green	TRUE	2025-10-22	2025-10-22T15:15:56+00:00	15:15:56+00:00	PT2H33M03S	5
	3	3	3.5	bluey	string	blue	FALSE	2025-10-22	2025-10-22T15:15:56+00:00	15:15:56+00:00	PT2H33M03S	10.5"""

    data_string = _as_delimited(string, delimiter)
    obj, metadata = read_csv_string_with_metadata(
        data_string, get_datamodel=get_datamodel, delimiter=delimiter
    )

    assert isinstance(obj, list)
    assert len(obj) == 3
    print("done")


@pytest.mark.parametrize("delimiter", ["\t", ","], ids=["tsv", "csv"])
def test_read_csv_string_with_metadata_transposed(delimiter):
    string = """#Title=ExampleArrayTransposed - HeaderDepth=3 - IsTransposed=True - DateTime=2025-10-22 15:42:29.557047 - DatamodelUrl=None
    #some hash string comment that should be ignored					
section	title	name	Column3	Column4	Column5
numeric	A Int	a_int	1	2	3
numeric	A Constrainedint	a_constrainedint	3	3	3
numeric	B Float	b_float	1.5	2.5	3.5
unicode	C Str	c_str	string	asdf	bluey
unicode	C Constrainedstr	c_constrainedstr	string	string	string
unicode	MyColor	d_enum	green	green	blue
boolean	E Bool	e_bool	TRUE	TRUE	FALSE
datetime	F Date	f_date	2025-10-22	2025-10-22	2025-10-22
datetime	G Datetime	g_datetime	2025-10-22T15:42:30+00:00	2025-10-22T15:42:30+00:00	2025-10-22T15:42:30+00:00
datetime	H Time	h_time	15:42:30+00:00	15:42:30+00:00	15:42:30+00:00
datetime	I Duration	i_duration	PT2H33M03S	PT2H33M03S	PT2H33M03S
numeric	B Calcfloat	b_calcfloat	1.5	5	10.5"""

    data_string = _as_delimited(string, delimiter)
    obj, metadata = read_csv_string_with_metadata(
        data_string, get_datamodel=get_datamodel, delimiter=delimiter
    )

    assert isinstance(obj, list)
    assert len(obj) == 3
    print("done")


def get_raw_jsonschema(metadata: DataGridMetaData) -> dict:
    return json.loads(PATH_JSONSCHEMA_RAW.read_text())


def test_read_excel_with_null(from_json_with_null):  # noqa: F811
    fpth, data, schema = from_json_with_null
    obj, metadata = read_excel_from_metadata(
        fpth, get_datamodel=lambda *args: schema.model_json_schema()
    )
    assert obj == data


def test_timedelta():
    # https://github.com/koxudaxi/datamodel-code-generator/issues/1624
    schema = {
        "title": "Test",
        "type": "object",
        "properties": {
            "a_int": {"default": 1, "title": "A Int", "type": "integer"},
            "i_duration": {
                "default": "PT2H33M3S",
                "format": "duration",
                "title": "I Duration",
                "type": "string",
            },
        },
    }

    Model = pydantic_model_from_json_schema(schema)
    assert (
        Model.model_fields["i_duration"].annotation == ty.Optional[datetime.timedelta]
    )


def test_field_name_change():
    """Test that the field name changes in the pydantic model if enumeration
    given same name as property.
    This behaviour introduced in version 0.28.5.
    Issue discussing this behaviour: https://github.com/koxudaxi/datamodel-code-generator/issues/2364
    """
    schema = {
        "type": "object",
        "properties": {"Fruit": {"$ref": "#/definitions/Fruit"}},
        "definitions": {
            "Fruit": {
                "enum": ["apple", "banana"],
                "name": "Fruit",
                "title": "Fruit",
                "type": "string",
            }
        },
    }
    test_model = pydantic_model_from_json_schema(schema)
    assert "Fruit_1" in test_model.__annotations__
    assert "Fruit" in test_model.model_fields["Fruit_1"].alias
    validated_model = test_model.model_validate({"Fruit": "apple"})
    assert validated_model.model_dump(by_alias=True, mode="json") == {"Fruit": "apple"}



import os
import pathlib

import pytest

FDIR_EXAMPLES = pathlib.Path(__file__).parent / "xlsxwriter-examples"


@pytest.fixture()
def move_dir():
    FDIR_EXAMPLES.mkdir(exist_ok=True)
    os.chdir(FDIR_EXAMPLES)


@pytest.mark.skip()
def test_write_chart(move_dir):
    #######################################################################
    #
    # An example of creating Excel Line charts with Python and XlsxWriter.
    #
    # SPDX-License-Identifier: BSD-2-Clause
    # Copyright 2013-2024, John McNamara, jmcnamara@cpan.org
    #
    import xlsxwriter

    workbook = xlsxwriter.Workbook("chart_line.xlsx")
    worksheet = workbook.add_worksheet()
    bold = workbook.add_format({"bold": 1})

    # Add the worksheet data that the charts will refer to.
    headings = ["Number", "Batch 1", "Batch 2"]
    data = [
        [2, 3, 4, 5, 6, 7],
        [10, 40, 50, 20, 10, 50],
        [30, 60, 70, 50, 40, 30],
    ]

    worksheet.write_row("A1", headings, bold)
    worksheet.write_column("A2", data[0])
    worksheet.write_column("B2", data[1])
    worksheet.write_column("C2", data[2])

    # Create a new chart object. In this case an embedded chart.
    chart1 = workbook.add_chart({"type": "line"})

    # Configure the first series.
    chart1.add_series(
        {
            "name": "=$B$1",
            "categories": "=Sheet1!$A$2:$A$7",
            "values": "=Sheet1!$B$2:$B$7",
        }
    )

    # Add a chart title and some axis labels.
    chart1.set_title({"name": "Results of sample analysis"})
    chart1.set_x_axis({"name": "Test number"})
    chart1.set_y_axis({"name": "Sample length (mm)"})

    # Set an Excel chart style. Colors with white outline and shadow.
    chart1.set_style(10)

    # Insert the chart into the worksheet (with an offset).
    worksheet.insert_chart("D2", chart1, {"x_offset": 25, "y_offset": 10})

    #######################################################################

    workbook.close()


@pytest.mark.skip()
def test_protected(move_dir):
    import xlsxwriter

    workbook = xlsxwriter.Workbook("protection.xlsx")
    worksheet = workbook.add_worksheet()

    # Create some cell formats with protection properties.
    unlocked = workbook.add_format({"locked": False})
    hidden = workbook.add_format({"hidden": True})

    # Format the columns to make the text more visible.
    worksheet.set_column("A:A", 40)

    # Turn worksheet protection on.
    worksheet.protect()

    # Write a locked, unlocked and hidden cell.
    worksheet.write("A1", "Cell B1 is locked. It cannot be edited.")
    worksheet.write("A2", "Cell B2 is unlocked. It can be edited.")
    worksheet.write("A3", "Cell B3 is hidden. The formula isn't visible.")

    worksheet.write_formula("B1", "=1+2")  # Locked by default.
    worksheet.write_formula("B2", "=1+2", unlocked)
    worksheet.write_formula("B3", "=1+2", hidden)

    workbook.close()



# std libs
import csv
import importlib.util
import json
import sys
import typing as ty
from datetime import date, datetime, time, timedelta, timezone
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory

from casefy import snakecase
from datamodel_code_generator import DataModelType, InputFileType, generate
from pydantic import AwareDatetime, BaseModel, ValidationError

# 3rd party
from python_calamine import CalamineSheet, CalamineWorkbook

# local
from xlsxdatagrid.xlsxdatagrid import DataGridMetaData


def _replace_empty_with_none(value: str) -> ty.Optional[str]:
    """Helper to replace empty strings with None."""
    return None if value == "" else value


def pydantic_model_from_json_schema(json_schema: str) -> ty.Type[BaseModel]:
    load = json_schema["title"].replace(" ", "") if "title" in json_schema else "Model"
    # TODO: refactor this when title vs name vs code has been sorted out...

    with TemporaryDirectory() as temporary_directory_name:
        temporary_directory = Path(temporary_directory_name)
        file_path = "model.py"
        module_name = file_path.split(".")[0]
        output = Path(temporary_directory / file_path)
        generate(
            json.dumps(json_schema),
            input_file_type=InputFileType.JsonSchema,
            input_filename="example.json",
            output=output,
            output_model_type=DataModelType.PydanticV2BaseModel,
            capitalise_enum_members=True,
        )
        spec = importlib.util.spec_from_file_location(module_name, output)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    return getattr(module, load)


def read_metadata(s: str) -> DataGridMetaData:
    s = s.replace("#", "")
    li = [x.split("=") for x in s.split(" - ")]
    di = {snakecase(x[0]): x[1] for x in li}
    return DataGridMetaData(**di)


def process_data(
    data: list[list[int | float | str | bool | time | date | datetime | timedelta],],
    is_transposed: bool = False,
    header_depth: int = 1,
    empty_string_to_none=True,
) -> list[dict]:
    if is_transposed:
        data = list(map(list, zip(*data)))

    header_names = [d[0] for d in data[0:header_depth]]

    if empty_string_to_none:
        data = [[_replace_empty_with_none(v) for v in row] for row in data]

    headers = {h: data[n] for n, h in enumerate(header_names)}
    header = headers[header_names[-1]]

    data = data[len(header_names) :]
    data = [dict(zip(header, d)) for d in data]

    return data


def process_data_with_metadata(
    data: list[list[int | float | str | bool | time | date | datetime | timedelta],],
    get_datamodel: ty.Optional[
        ty.Callable[[DataGridMetaData], ty.Type[BaseModel]]
    ] = None,
) -> tuple[list[dict], DataGridMetaData]:
    metadata = read_metadata(data[0][0])
    data_without_comments = drop_leading_comments(data)
    processed_data = process_data(
        data_without_comments,
        metadata.is_transposed,
        metadata.header_depth,
        True,
    )
    processed_metadata = process_metadata(metadata, data)
    json_schema = (
        get_datamodel(processed_metadata) if get_datamodel is not None else None
    )
    if json_schema is not None:
        pydantic_model = pydantic_model_from_json_schema(json_schema)
    if pydantic_model is not None:
        processed_data = pydantic_validate_data(processed_data, pydantic_model)
    return processed_data, processed_metadata


def process_metadata(metadata: DataGridMetaData, data: list[dict]) -> DataGridMetaData:
    """Extracts DataGridMetaData from the given data."""
    hd = metadata.header_depth
    is_t = metadata.is_transposed
    if is_t:
        data = list(map(list, zip(*data)))

    header_names = [d[0] for d in data[0:hd]]
    data = [d[1:] for d in data]
    headers = {h: data[n] for n, h in enumerate(header_names)}
    metadata.datagrid_index_name = list(headers.keys())
    metadata.header = list(headers.values())

    return metadata


def get_datamodel(metadata: DataGridMetaData) -> dict:
    pass


def make_datetime_tz_aware(data, pydantic_model):
    def field_is_aware_datetime(field):
        if hasattr(field.annotation, "__args__"):
            if AwareDatetime in field.annotation.__args__:
                return True
            else:
                return False
        elif isinstance(field.annotation, AwareDatetime):
            return True
        else:
            return False

    row_model = pydantic_model.model_fields["root"].annotation.__args__[0]
    keys = [k for k, v in row_model.model_fields.items() if field_is_aware_datetime(v)]
    if len(keys) > 0:
        # HACK: if data is from csv string, run the pydantic validation before adding timezone info
        if type(data[0][keys[0]]) is str:
            data = pydantic_model.model_validate(data).model_dump(
                mode="python", by_alias=True
            )
        return [d | {k: d[k].replace(tzinfo=timezone.utc) for k in keys} for d in data]
    else:
        return data


def drop_leading_comments(data: list[list[str]]) -> list[list[str]]:
    """
    Removes only the initial consecutive lines that start with '#'
    from the list of CSV rows.
    """
    for i, row in enumerate(data):
        # skip empty rows too
        if not row or not row[0].lstrip().startswith("#"):
            return data[i:]  # return from the first non-comment line onward
    return []


def get_list_of_list_from_worksheet(worksheet: CalamineSheet) -> list[list]:
    return worksheet.to_python(skip_empty_area=True)


def get_list_of_list_from_string(
    csv_string: str, delimiter: str = "\t"
) -> list[list[int | float | str | bool | time | date | datetime | timedelta],]:
    csv_file = StringIO(csv_string.strip())
    reader = csv.reader(csv_file, delimiter=delimiter)
    data = [x for x in reader]
    return data


def pydantic_validate_data(
    data, pydantic_model: BaseModel, return_pydantic_model: bool = False
):
    data = make_datetime_tz_aware(data, pydantic_model)
    obj = pydantic_model.model_validate(data)
    if return_pydantic_model:
        return obj
    else:
        return obj.model_dump(mode="json", by_alias=True)


def read_excel_from_metadata(
    path,
    get_datamodel: ty.Optional[
        ty.Callable[[DataGridMetaData], ty.Type[BaseModel]]
    ] = None,
):
    workbook = CalamineWorkbook.from_path(path)
    sheet = workbook.sheet_names[0]
    worksheet = workbook.get_sheet_by_name(sheet)
    data = get_list_of_list_from_worksheet(worksheet)
    return process_data_with_metadata(data, get_datamodel)


def read_list_of_lists(
    data: list[list[int | float | str | bool | time | date | datetime | timedelta],],
    is_transposed: bool = False,
    header_depth: int = 1,
    model: BaseModel | None = None,
) -> list[dict]:
    data = drop_leading_comments(data)
    processed_data = process_data(data, is_transposed, header_depth, True)
    try:
        validated_data = pydantic_validate_data(processed_data, model)
        return validated_data, []
    except ValidationError as exc:
        return [], exc.errors()
    # TODO @Arshadwaqas115: why returning a tuple? better to return data only


def read_excel(
    path,
    is_transposed: bool = False,
    header_depth: int = 1,
    model: BaseModel | None = None,
):
    workbook = CalamineWorkbook.from_path(path)
    sheet = workbook.sheet_names[0]
    worksheet = workbook.get_sheet_by_name(sheet)
    data = get_list_of_list_from_worksheet(worksheet)
    return read_list_of_lists(data, is_transposed, header_depth, model)


def read_csv_string(
    csv_string: str,
    is_transposed: bool = False,
    header_depth: int = 1,
    model: BaseModel | None = None,
    delimiter: str = ",",
) -> list[dict]:
    """Read a CSV string and process it into a list of dicts, optionally validating with a pydantic model. Note: This gets rid of all initial comments in the CSV string before the data begins.

    Args:
        csv_string (str): A csv string (could be a tsv string if delimiter is set to "\t").
        is_transposed (bool, optional): Configure whether the csv string is transposed or not. Defaults to False.
        header_depth (int, optional): The header depth in the csv string. Defaults to 1.
        model (BaseModel | None, optional): The pydantic model. Defaults to None.
        delimiter (str, optional): The delimiter that the data is separated on. Defaults to ",".
        previous_value (list[dict] | None, optional): . Defaults to None.

    Returns:
        _type_: List of dictionaries representing the data rows.
    """

    data = get_list_of_list_from_string(csv_string, delimiter=delimiter)
    return read_list_of_lists(data, is_transposed, header_depth, model)


def read_csv_string_with_metadata(
    csv_string: str,
    get_datamodel: ty.Optional[
        ty.Callable[[DataGridMetaData], ty.Type[BaseModel]]
    ] = None,
    delimiter: str = ",",
):
    data = get_list_of_list_from_string(csv_string, delimiter=delimiter)
    return process_data_with_metadata(data, get_datamodel)



import functools
import inspect
import logging
import pathlib
import typing as ty
from datetime import date, datetime, time, timedelta

import annotated_types
import pandas as pd
import xlsxwriter as xw
from annotated_types import doc
from dirty_equals import IsInstance
from jsonref import replace_refs
from pandas.io.json import build_table_schema
from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    HttpUrl,
    ImportString,
    RootModel,
    computed_field,
    model_validator,
)
from pydantic_extra_types.color import Color
from typing_extensions import Annotated, Self
from xlsxwriter.utility import _datetime_to_excel_datetime, xl_rowcol_to_cell

from xlsxdatagrid.colours import get_color_pallette

# https://specs.frictionlessdata.io//table-schema/
name_doc = """The field descriptor MUST contain a name property.
 This property SHOULD correspond to the name of field/column in the data file (if it has a name).
 As such it SHOULD be unique (though it is possible, but very bad practice, for the data file to have multiple columns with the same name). 
 name SHOULD NOT be considered case sensitive in determining uniqueness. 
 However, since it should correspond to the name of the field in the data file it may be important to preserve case."""


# NOT IN USE -------------------------------
class HeaderStyling(BaseModel):  # matches ipydatagrid
    header_background_color: ty.Optional[Color] = Field(
        None, description="background color for all non-body cells (index and columns)"
    )
    header_grid_line_color: ty.Optional[Color] = Field(
        None, description="grid line color for all non-body cells (index and columns)"
    )
    header_vertical_grid_line_color: ty.Optional[Color] = Field(
        None, description="vertical grid line color for all non-body cells"
    )
    header_horizontal_grid_line_color: ty.Optional[Color] = Field(
        None, description="horizontal grid line color for all non-body cells"
    )
    header_selection_fill_color: ty.Optional[Color] = Field(
        None,
        description="fill color of headers intersecting with selected area at column or row",
    )
    header_selection_border_color: ty.Optional[Color] = Field(
        None,
        description="border color of headers intersecting with selected area at column or row",
    )


XL_FORMAT_PROPERTIES = (
    "font_name",
    "font_size",
    "font_color",
    "bold",
    "italic",
    "underline",
    "font_strikeout",
    "font_script",
    "num_format",
    "locked",
    "hidden",
    "align",
    "valign",
    "rotation",
    "text_wrap",
    "reading_order",
    "text_justlast",
    "center_across",
    "indent",
    "shrink",
    "pattern",
    "bg_color",
    "fg_color",
    "border",
    "bottom",
    "top",
    "left",
    "right",
    "border_color",
    "bottom_color",
    "top_color",
    "left_color",
    "right_color",
)
# ^these are set at property level per column

XL_TABLE_PROPERTIES = (
    "data",
    "autofilter",
    "header_row",
    "banded_columns",
    "banded_rows",
    "first_column",
    "last_column",
    "style",
    "total_row",
    "columns",
    "name",
)
# ^these are set at schema level for the whole table

# ^ NOT IN USE. COULD USE FOR VALIDATION... ----------------------


METADATA_FSTRING: str = "#Title={title} - HeaderDepth={header_depth} - IsTransposed={is_transposed} - DateTime={now} - DatamodelUrl={datamodel_url}"
# ^ TODO: build a metadrata string from what data is present. allow additions to this string but not removals.

MAP_TYPES_JSON_XL = {"integer": "integer", "float": "decimal", "date": "date"}


XL_TABLE_COLUMNS_PROPERTIES = (
    "header",
    "header_format",
    "formula",
    "total_string",
    "total_function",
    "total_value",
    "format",
)


DATETIME_STR = 'yyyy-mm-ddThh:mm:ss"+00:00"'
DATE_STR = "yyyy-mm-dd"
TIME_STR = 'hh:mm:ss"+00:00"'
DURATION_STR = '"PT"h"H"mm"M"ss"S"'

DATETIME_FORMAT = {"num_format": DATETIME_STR}
DATE_FORMAT = {"num_format": DATE_STR}
TIME_FORMAT = {"num_format": TIME_STR}
DURATION_FORMAT = {"num_format": DURATION_STR}

PY2XL = {
    "**": "^",
    "!=": "<>",
    # other simple arithemetic operators the same
}


class Constraints(BaseModel):
    minimum: ty.Optional[ty.Union[int, float]] = None
    maximum: ty.Optional[ty.Union[int, float]] = None
    exclusiveMinimum: ty.Optional[bool] = None
    exclusiveMaximum: ty.Optional[bool] = None
    enum: ty.Optional[list[ty.Any]] = None
    maxLength: ty.Optional[int] = None
    minLength: ty.Optional[int] = None


NUMERIC_CONSTRAINTS = [
    i for i in list(Constraints.__annotations__.keys()) if i != "enum"
]
LI_CONSTRAINTS = list(Constraints.__annotations__.keys())
# https://xlsxwriter.readthedocs.io/working_with_data_validation.html#criteria


def py2xl_formula(formula, map_names):
    def replace(formula, di):
        for k, v in di.items():
            if k in formula:
                formula = formula.replace(k, v)
        return formula

    map_table_names = {x: f"[@[{x}]]" for x in map_names.keys()}
    formula = replace(formula, PY2XL)
    formula = replace(formula, map_names)
    formula = replace(formula, map_table_names)
    return "= " + formula


def get_numeric_constraints(di):
    return [k for k in di.keys() if k in NUMERIC_CONSTRAINTS]


def map_simple_numeric_constraints(di):
    pass


def map_constraints(di):
    get_numeric_constraints(di)
    return None


class FieldSchema(BaseModel):
    model_config = ConfigDict(extra="allow")

    name: Annotated[str, doc(name_doc)]
    type: str  # TODO: create enum
    format: str = None
    title: Annotated[str, doc("A human readable label or title for the field")] = None
    description: Annotated[
        str, doc("A description for this field e.g. 'The recipient of the funds'")
    ] = None
    example: Annotated[str, doc("An example value for the field")] = None
    constraints: ty.Optional[Constraints] = (
        None  # this is what the https://specs.frictionlessdata.io//table-schema/ but other things
    )
    formula: ty.Optional[str] = None


class FieldSchemaXl(FieldSchema):
    data_validation: dict = {}
    conditional_format: str
    cell_format: dict
    xl_formula: ty.Optional[str] = None


def get_xl_constraints(f: FieldSchema):  # TODO: write text for this
    if f.type == "boolean":
        return {
            "validate": "list",
            "source": ["TRUE", "FALSE"],
            "input_title": "must be a boolean value",
            "input_message": "TRUE or FALSE only",
        }
    if f.constraints is None:
        return None
    if f.constraints.enum is not None:
        return {
            "validate": "list",
            "source": f.constraints.enum,
            "input_title": "select from list",
            "input_message": f"must be one of: {str(f.constraints.enum)}",
        }

    if f.type == "integer" or f.type == "number" or f.type == "date":
        _type = "decimal" if f.type == "number" else f.type
        _min, _max, _exmin, _exmax = (
            f.constraints.minimum,
            f.constraints.maximum,
            f.constraints.exclusiveMinimum,
            f.constraints.exclusiveMaximum,
        )
        if _min is not None and _max is not None:
            return {
                "validate": _type,
                "criteria": "between",
                "minimum": _min,
                "maximum": _max,
                "input_title": f"select {_type}",
                "input_message": f"between {_min} and {_max}",
            }
        elif _min is not None and _exmin is not None:
            return {
                "validate": _type,
                "criteria": "<",
                "minimum": _min,
                "input_title": f"select {_type}",
                "input_message": f"exclusive minimum {_min}",
            }
        elif _min is not None and _exmin is None:
            return {
                "validate": _type,
                "criteria": "=<",
                "minimum": _min,
                "input_title": f"select {_type}",
                "input_message": f"minimum {_min}",
            }
        elif _max is not None and _exmax is not None:
            return {
                "validate": _type,
                "criteria": ">",
                "maximum": _max,
                "input_title": f"select {_type}",
                "input_message": f"exclusive maximum {_max}",
            }
        elif _max is not None and _exmax is None:
            return {
                "validate": _type,
                "criteria": ">=",
                "maximum": _max,
                "input_title": f"select {_type}",
                "input_message": f"maximum {_max}",
            }
        else:
            ValueError(
                f"no numeric match found for: min={_min}, max={_max}, type={_type}"
            )
    if f.type == "string" and f.constraints.maxLength is not None:
        return {
            "validate": "length",
            "criteria": "<",
            "value": f.constraints.maxLength,
            "input_title": f"select {f.type}",
            "input_message": f"with less than {f.constraints.maxLength} characters",
        }
    if f.type == "string" and f.constraints.minLength is not None:
        return {
            "validate": "length",
            "criteria": ">",
            "value": f.constraints.minLength,
            "input_title": f"select {f.type}",
            "input_message": f"with less than {f.constraints.minLength} characters",
        }


class DataGridMetaData(BaseModel):
    model_config = ConfigDict(exclude_none=True)
    title: str = Field(validation_alias=AliasChoices("title", "Title"))
    name: ty.Optional[str] = Field(
        None, validation_alias=AliasChoices("template_name", "name")
    )
    is_transposed: bool = False  # TODO: rename -> display_transposed
    header_depth: int = Field(1, validate_default=True)
    # include_header_titles: bool = True  # TODO
    datamodel_url: ty.Optional[HttpUrl] = Field(
        None, validation_alias=AliasChoices("datamodel_url", "DatamodelUrl")
    )
    datamodel_path: ty.Optional[pathlib.Path] = Field(
        None, validation_alias=AliasChoices("datamodel_path", "DatamodelPath")
    )  # TODO: add as an option
    datamodel_importstr: ty.Optional[ImportString] = Field(
        None, validation_alias=AliasChoices("datamodel_importstr", "DatamodelImportstr")
    )  # TODO: add as an option. preferred when present.
    metadata_fstring: str = Field(
        METADATA_FSTRING
    )  # TODO: should this be fixed... or validate that the base string is included...
    date_time: ty.Optional[datetime] = None
    datagrid_index_name: tuple = ("name",)  # RENAME: header_field_keys
    header: list[list[str]] = []

    @model_validator(mode="before")
    @classmethod
    def check_datamodel_url(cls, data: ty.Any) -> ty.Any:
        if isinstance(data, dict):
            if "datamodel_url" in data:
                if data["datamodel_url"] == "None":
                    data["datamodel_url"] = None
        return data

    @computed_field
    def now(self) -> datetime:
        return datetime.now()

    @model_validator(mode="after")
    def check_name(self) -> Self:
        if self.name is None:
            self.name = self.title.replace(" ", "")
        return self


class DataGridSchema(DataGridMetaData):
    model_config = ConfigDict(extra="allow")

    header_background_color: ty.Optional[Color] = None
    base_row_size: int = 20
    base_column_size: int = 64
    base_row_header_size: int = 64
    base_column_header_size: int = 20
    column_widths: dict[str, float] = {}
    fields: list[FieldSchema]

    @model_validator(mode="after")
    def get_header_depth(self) -> "DataGridSchema":
        self.header_depth = len(self.datagrid_index_name)
        self.header = [
            [
                (lambda f, nm: getattr(f, nm) if hasattr(f, nm) else None)(f, nm)
                for f in self.fields
            ]
            for nm in self.datagrid_index_name
        ]
        return self

    @computed_field
    def map_name_header(self) -> dict[str, ty.Union[str, list[str]]]:
        if self.datagrid_index_name == ("name",):
            return None
        elif len(self.datagrid_index_name) == 1:
            return {
                f.name: getattr(f, self.datagrid_index_name[0]) for f in self.fields
            }
        elif len(self.datagrid_index_name) > 1:
            return {
                f.name: [getattr(f, nm) for nm in self.datagrid_index_name]
                for f in self.fields
            }
        else:
            raise ValueError("incorrect type")

    @computed_field
    def field_names(self) -> list:
        return [f.name for f in self.fields]


def convert_date_to_excel_ordinal(d: date, offset: int = 693594):
    # the offset date value for the date of 1900-01-00 = 693594
    return d.toordinal() - offset


def get_datetime(d):
    return RootModel[datetime](d).model_dump()


def get_time(d):
    return RootModel[time](d).model_dump()


def get_date(d):
    return RootModel[date](d).model_dump()


def get_duration(d):
    return RootModel[timedelta](d).model_dump()


class XlGrid(BaseModel):
    header_sections: list = ["section", "category"]  # used to colour code only
    xy: tuple[int, int] = 0, 0  # row, col
    xy_arrays: dict[str, tuple[int, int]] = {"a": (0, 0)}
    format_arrays: dict[str, str] = Field({})  # col-name: format
    comment_arrays: dict[str, str] = Field({})  # col-name: comment
    rng_arrays: dict[str, tuple[int, int, int, int]] = {"a": (0, 0, 0, 0)}
    xy_headers: list[tuple[int, int]] = [(0, 0)]
    rng_headers: list[tuple[int, int, int, int]] = [(0, 0, 0, 0)]
    format_headers: list = [None]
    tbl_range: tuple[int, int, int, int] = (0, 0, 0, 0)
    tbl_headers: ty.Optional[list] = None
    validation_arrays: ty.Optional[dict[str, dict]] = (
        None  # add validation to simple types: integer, float, string
    )
    formula_arrays: dict[str, str] = {}
    formats: dict[str, dict] = {
        "datetime": DATETIME_FORMAT,
        "date": DATE_FORMAT,
        "time": TIME_FORMAT,
        "duration": DURATION_FORMAT,
    }
    conditional_formats: list[tuple[int, int, int, int, dict]] = []
    hide_gridlines: Annotated[int, annotated_types.Interval(ge=0, le=2)] = (
        2  # hidden by default
    )
    metadata: str = ""
    # length: int = 0


class XlTableWriter(XlGrid):
    gridschema: DataGridSchema
    data: dict[str, list]


class DataGridData(RootModel):
    root: dict[ty.Union[str, int, float], list]  # columnar data


def generate_metadata_string(gridschema: DataGridSchema) -> str:
    metadata = gridschema.metadata_fstring.format(**gridschema.model_dump())
    return metadata


def get_xlgrid(
    gridschema: DataGridSchema,
    data: DataGridData,
    exclude_metadata: bool = False,
) -> XlGrid:
    data = data.root
    header_sections = ["section", "category"]
    start_coordinates = 0, 0
    formats = {
        "datetime": DATETIME_FORMAT,
        "date": DATE_FORMAT,
        "time": TIME_FORMAT,
        "duration": DURATION_FORMAT,
    }
    format_arrays = {}
    # ^ ?
    metadata = generate_metadata_string(gridschema)

    is_t = gridschema.is_transposed
    ix_nm = gridschema.datagrid_index_name  # column headings
    x, y = start_coordinates
    hd = gridschema.header_depth  # header depth
    fd_nns = gridschema.field_names  # field names
    length = len(data[fd_nns[0]]) - 1 if len(data) > 0 else 0  # length of data arrays
    format_headers = hd * [None]

    di = {}
    di["metadata"] = metadata
    di["format_headers"] = format_headers

    # ensure data and key col names in same order - TODO: check. shouldn't modify data here...
    if gridschema.field_names != list(data.keys()):
        data = {
            x: (lambda x, data: data[x] if x in data.keys() else [None] * length)(
                x, data
            )
            for x in gridschema.field_names
        }
    assert gridschema.field_names == list(data.keys())
    # TODO: allow option of only outputting fields which have data associated with them

    if is_t:
        x += 1
    # ^ leave room for the header names

    x += 1 if not exclude_metadata else 0
    # ^ leave room for the metadata

    if not is_t:  # build a normal xl table
        xy_arrays = {
            f: (x + hd, y + n) for n, f in enumerate(fd_nns)
        }  # +1 as below header row
        rng_arrays = {
            k: (v[0], v[1], v[0] + length, v[1]) for k, v in xy_arrays.items()
        }
        beg, end = rng_arrays[fd_nns[0]][0:2], rng_arrays[fd_nns[-1]][2:4]
        beg = (beg[0] - 1, beg[1])  # inc. header
        tbl_range = (*beg, *end)
        di["tbl_range"] = tbl_range

        xy_headers = [(x + n, y) for n in range(0, hd)]
        rng_headers = [
            (v[0], v[1], v[0], v[1] + len(xy_arrays)) for v in xy_headers
        ]  # (x1, y1, x2, y2)

        tbl_headers = gridschema.header[-1]

        # formulas currently only posible with normal tables
        map_names = {
            k: v for k, v in zip(set(gridschema.field_names), set(tbl_headers))
        }
        formula_arrays = {
            f.name: py2xl_formula(f.formula, map_names)
            for f in gridschema.fields
            if f.formula is not None
        } | {
            f.name: getattr(f, "xl_formula")
            for f in gridschema.fields
            if hasattr(f, "xl_formula") and getattr(f, "xl_formula") is not None
        }  # formula override (normally empty)

    else:  # build a transposed xl table
        xy_arrays = {
            f: (x + n, y + hd) for n, f in enumerate(fd_nns)
        }  # +1 as below header row
        rng_arrays = {
            k: (v[0], v[1], v[0], v[1] + length) for k, v in xy_arrays.items()
        }
        beg, end = rng_arrays[fd_nns[0]][0:2], rng_arrays[fd_nns[-1]][2:4]
        beg = (beg[0] - 1, beg[1] - hd)  # inc. header
        tbl_range = (*beg, *end)
        xy_headers = [(x, y + n) for n in range(0, hd)]
        rng_headers = [
            (v[0], v[1], v[0] + len(xy_arrays), v[1]) for v in xy_headers
        ]  # (x1, y1, x2, y2)
        tbl_headers = None
        formula_arrays = {}

    # get colour the header sections
    palettes_in_use = []
    header_sections = [h for h in header_sections if h in ix_nm]
    di_section_colors = {}
    for h in header_sections:
        sections = list(set([getattr(f, h) for f in gridschema.fields]))
        colors = get_color_pallette(len(sections), palettes_in_use)
        di_section_colors = di_section_colors | {
            h: [(s, c) for s, c in zip(sections, colors)]
        }

    for v in di_section_colors.values():
        formats = formats | {f"{x[1]}": {"bg_color": x[1]} for x in v}

    # colour the header sections
    conditional_formats = []
    for k, v in di_section_colors.items():
        r1, c1, r2, c2 = rng_headers[ix_nm.index(k)]
        for _ in v:
            conditional_formats = conditional_formats + [
                [
                    r1,
                    c1,
                    r2,
                    c2,
                    {
                        "type": "cell",
                        "criteria": "equal to",
                        "value": '"{0}"'.format(_[0]),
                        "format": _[1],  # formats[_[1]]
                    },
                ]
            ]

    validation_arrays = {f.name: get_xl_constraints(f) for f in gridschema.fields}
    validation_arrays = {k: v for k, v in validation_arrays.items() if v is not None}
    dates = [f.name for f in gridschema.fields if f.format == "date"]
    date_times = [f.name for f in gridschema.fields if f.format == "date-time"]
    durations = [f.name for f in gridschema.fields if f.format == "duration"]
    times = [f.name for f in gridschema.fields if f.format == "time"]

    for d in date_times:
        data[d] = [
            _datetime_to_excel_datetime(get_datetime(v), False, True) for v in data[d]
        ]
        format_arrays[d] = "datetime"
    for d in dates:
        data[d] = [
            _datetime_to_excel_datetime(get_datetime(v), False, True) for v in data[d]
        ]
        format_arrays[d] = "date"
    for d in times:
        data[d] = [
            _datetime_to_excel_datetime(get_time(v), False, True) for v in data[d]
        ]
        format_arrays[d] = "time"
    for d in durations:
        data[d] = [
            _datetime_to_excel_datetime(get_duration(v), False, True) for v in data[d]
        ]
        format_arrays[d] = "duration"

    di = {}

    di["xy"] = start_coordinates
    di["header_sections"] = header_sections
    di["metadata"] = metadata
    di["formats"] = formats
    di["format_headers"] = format_headers
    di["formula_arrays"] = formula_arrays
    di["format_arrays"] = format_arrays
    di["xy_headers"] = xy_headers
    di["xy_arrays"] = xy_arrays
    di["rng_headers"] = rng_headers
    di["rng_arrays"] = rng_arrays
    di["tbl_range"] = tbl_range
    di["tbl_headers"] = tbl_headers
    di["validation_arrays"] = validation_arrays
    di["conditional_formats"] = conditional_formats
    di["comment_arrays"] = {}
    di["hide_gridlines"] = 2  # hidden by default

    return XlGrid(**di)


def flatten_allOf(di: dict) -> dict:
    if "allOf" in di.keys():
        for _ in di["allOf"]:
            di = {**di, **_}
        return {k: v for k, v in di.items() if k != "allOf"}
    else:
        return di


def flatten_anyOf(fields: list) -> list:
    for field in fields:
        if "anyOf" in field.keys():
            types = list(set([f["type"] for f in field["anyOf"]]))
            if len(types) == 2:
                field["type"] = field["anyOf"][0]["type"]
            if len(types) > 2:
                logging.warning(
                    f"more than 2 types allowed ({types})... for {field['name']}"
                )
            field.pop("anyOf")
    return fields


def convert_records_to_datagrid_schema(schema: dict):
    li_constraints = list(Constraints.__annotations__.keys())
    gridschema = replace_refs(schema, merge_props=True)
    gridschema["fields"] = [
        flatten_allOf(v) | {"name": k}
        for k, v in gridschema["items"]["properties"].items()
    ]
    gridschema["fields"] = flatten_anyOf(gridschema["fields"])
    # move constraints
    for n in range(0, len(gridschema["fields"])):
        for k in list(gridschema["fields"][n].keys()):
            if k in li_constraints:
                if "constraints" not in list(gridschema["fields"][n].keys()):
                    gridschema["fields"][n]["constraints"] = {}
                gridschema["fields"][n]["constraints"] = gridschema["fields"][n][
                    "constraints"
                ] | {k: gridschema["fields"][n][k]}
                gridschema["fields"][n].pop(k)

    gridschema["format"] = "dataframe"
    gridschema = {k: v for k, v in gridschema.items() if k not in ["$defs", "items"]}
    return gridschema


def convert_list_records_to_dict_arrays(data: list[dict]) -> dict[str, list]:
    if len(data) == 0:
        return {}
    return {k: [dic[k] if k in dic else None for dic in data] for k in data[0]}


def convert_dict_arrays_to_list_records(
    data: dict[str, list],
) -> list[dict]:  # USED IN TESTS ONLY
    if len(data) == 0:
        return []
    return [dict(zip(data.keys(), values)) for values in zip(*data.values())]


def coerce_data(
    data: ty.Union[dict[ty.Union[str, int, float], list], list[dict], pd.DataFrame],
) -> DataGridData:
    if isinstance(data, DataGridData):
        return data
    elif isinstance(data, dict):
        return DataGridData(root=data)
    elif isinstance(data, list):
        return DataGridData(root=convert_list_records_to_dict_arrays(data))
    elif isinstance(data, pd.DataFrame):
        return DataGridData(
            root=convert_list_records_to_dict_arrays(data.to_dict(orient="records"))
        )
    elif isinstance(data, RootModel):
        return DataGridData(root=convert_list_records_to_dict_arrays(data.model_dump()))
    else:
        raise ValueError(f"unrecognised input datatype {type(data)}")


def coerce_schema(
    schema: ty.Union[dict, DataGridSchema, BaseModel, ty.Type[BaseModel]],
) -> DataGridSchema:
    """pass various types of grid schema definitions and return a `DataGridSchema`"""
    if schema == IsInstance(DataGridSchema, only_direct_instance=True):
        return schema  # its already a `DataGridSchema`
    elif isinstance(schema, BaseModel):
        return DataGridSchema(
            **convert_records_to_datagrid_schema(schema.model_json_schema())
        )  # its a pydantic object
    elif inspect.isclass(schema) and issubclass(schema, BaseModel):
        return DataGridSchema(
            **convert_records_to_datagrid_schema(schema.model_json_schema())
        )  # its a pydantic model
    elif (
        isinstance(schema, dict)
        and "fields" in schema.keys()
        and isinstance(schema["fields"], list)
    ):
        return DataGridSchema(**schema)  # its a frictionless schema
    elif (
        isinstance(schema, dict)
        and schema["type"] == "array"
        and "items" in schema.keys()
    ):
        return DataGridSchema(
            **convert_records_to_datagrid_schema(schema)
        )  # its a json schema as array of object records
    else:
        raise AttributeError(
            "schema must be a `DataGridSchema`,",
            "a pydantic array of objects jsonschema",
            f"or a frictionless datagrid schema, not: {schema}",
        )


def write_grid(
    workbook,
    xlgrid: XlGrid,
    gridschema: DataGridSchema,
    data: DataGridData,
    exclude_metadata: bool = False,
) -> xw.worksheet.Worksheet:
    data = data.root
    name = gridschema.title
    worksheet = workbook.add_worksheet(name=name)
    header_depth = len(gridschema.datagrid_index_name)  # header depth
    write_array = (
        worksheet.write_row if gridschema.is_transposed else worksheet.write_column
    )
    write_header = (
        worksheet.write_row if not gridschema.is_transposed else worksheet.write_column
    )
    header_index = (
        xlgrid.xy_headers[-1][1]
        if gridschema.is_transposed
        else xlgrid.xy_headers[-1][0]
    )
    set_header_border = (
        functools.partial(worksheet.set_row, header_index)
        if not gridschema.is_transposed
        else functools.partial(worksheet.set_column, header_index, header_index)
    )
    freeze_panes = (
        (0, header_index + 1) if gridschema.is_transposed else (header_index + 1, 0)
    )
    header_border = {"right": 5} if gridschema.is_transposed else {"bottom": 5}
    formats = {k: workbook.add_format(v) for k, v in xlgrid.formats.items()}
    format_arrays = {k: formats[v] for k, v in xlgrid.format_arrays.items()}
    conditional_formats = []

    for c in xlgrid.conditional_formats:
        di = c[4]
        f = di["format"]
        di = di | {"format": formats[f]}
        conditional_formats += [list(c[0:4]) + [di]]
    _format = dict(bold=True) | header_border | {"locked": True}
    if gridschema.is_transposed:
        _format = _format | {"align": "right"}

    header_cell_format = workbook.add_format(_format)
    calc_cell_format = workbook.add_format(dict(font_color="blue", italic=True))
    header_label_cell_format = workbook.add_format(
        dict(font_color="#999999", italic=True)
    )
    header_white_cell_format = workbook.add_format(dict(font_color="#FFFFFF"))

    # special formats for arrays (mostly used for datetime)
    # for k, v in xlgrid.format_arrays.items():

    # make table --------------------------
    length = len(list(data.values())[0]) + len(gridschema.datagrid_index_name)

    def get_name(n, header_depth):
        return f"Column{n}" if n >= header_depth else gridschema.datagrid_index_name[n]

    column_labels = [get_name(n, header_depth) for n in range(0, length)]

    formula_columns = []
    if gridschema.is_transposed:  # transposed - with headers
        columns = [{"header": c} for c in column_labels]
        for idx in range(0, min(header_depth, len(columns))):
            header_value = columns[idx]["header"]
            if isinstance(header_value, str):
                columns[idx]["header"] = f"#{header_value}"

        options = dict(
            style="Table Style Light 1",
            header_row=True,
            first_column=False,
            columns=columns,
        )

    else:  # not transposed - with headers
        columns = {
            f.name: {"header": h}
            | {
                k: v
                for k, v in f.model_dump(exclude_none=True).items()
                if k in XL_TABLE_COLUMNS_PROPERTIES[0:6]
            }
            for h, f in zip(xlgrid.tbl_headers, gridschema.fields)
        }
        for k, v in xlgrid.formula_arrays.items():
            if "formula" in columns[k].keys():
                formula_columns += [k]
                columns[k]["formula"] = xlgrid.formula_arrays[k]
                columns[k]["format"] = calc_cell_format

        for k, v in xlgrid.format_arrays.items():
            columns[k]["format"] = formats[v]
        # ^ TODO: formatting dates and datetime as numeric with excel string formatting
        options = dict(
            style="Table Style Light 1",
            header_row=True,
            first_column=False,
            columns=list(columns.values()),
        )

    options = options  # | {"name": name}  # TODO: <- table name needs to not inc. spaces etc... update
    # ^ a known table name will be important if / when we want to do lookups between tables...

    worksheet.add_table(*xlgrid.tbl_range, options)
    # NOTE: if you write a table to excel with a header - the table range includes the header.

    # -------------------------------------
    # write arrays
    for k, v in xlgrid.xy_arrays.items():
        if k not in formula_columns:
            # li = list(np.array(data[k]))
            li = data[k]

            if k in format_arrays:
                write_array(*v, li, format_arrays[k])
            else:
                write_array(*v, li)

    if len(gridschema.datagrid_index_name) > 0:
        if gridschema.is_transposed:
            rngs = xlgrid.xy_headers
        else:
            rngs = xlgrid.xy_headers[:-1]
            # table headers are inserted into when adding table
        {
            write_header(*start_cell, data, header_cell_format)
            for start_cell, data in zip(rngs, gridschema.header)
        }

    if xlgrid.validation_arrays is not None:
        for k, v in xlgrid.validation_arrays.items():
            if v is not None:
                rng = xlgrid.rng_arrays[k]
                worksheet.data_validation(*rng, options=v)
    # print("add conditional formats...")
    for c in conditional_formats:
        worksheet.conditional_format(*c)

    # apply header border
    cell_format = workbook.add_format(dict(valign="top") | header_border)
    set_header_border(None, cell_format)

    # write column labels
    x, y = xlgrid.xy
    x += 1 if not exclude_metadata else 0  # for metadata row
    label_values = column_labels[0:header_depth]
    if gridschema.is_transposed:
        label_values = [
            f"#{label}" if isinstance(label, str) else label for label in label_values
        ]
    if gridschema.is_transposed:
        # set empty table headers to be white
        y += header_depth
        write_array(
            *(x, y),
            column_labels[header_depth : len(column_labels)],
            header_white_cell_format,
        )

    # write array comments
    for k, v in xlgrid.comment_arrays.items():
        cell = xlgrid.xy_arrays[k]
        worksheet.write_comment(xl_rowcol_to_cell(*cell), *v)

    worksheet.freeze_panes(*freeze_panes)
    worksheet.autofit()
    worksheet.hide_gridlines(xlgrid.hide_gridlines)
    # write metadata
    if not exclude_metadata:
        worksheet.write(*xlgrid.xy, xlgrid.metadata, header_label_cell_format)
    return worksheet


def write_sheet(
    workbook: xw.Workbook,
    data: list[dict],
    gridschema: ty.Union[dict, DataGridSchema, BaseModel, ty.Type[BaseModel]],
    exclude_metadata: bool = False,
) -> tuple[xw.worksheet.Worksheet, XlTableWriter]:
    gridschema = coerce_schema(gridschema)
    griddata = coerce_data(data)
    xlgrid = get_xlgrid(
        gridschema,
        griddata,
        exclude_metadata=exclude_metadata,
    )
    wsheet = write_grid(
        workbook,
        xlgrid,
        gridschema,
        griddata,
        exclude_metadata=exclude_metadata,
    )

    xl_tbl = XlTableWriter(
        data=griddata.model_dump(), gridschema=gridschema, **xlgrid.model_dump()
    )
    # data = convert_list_records_to_dict_arrays(data)
    return wsheet, xl_tbl


def write_sheets(
    workbook: xw.Workbook,
    datas: list[list[dict]],
    gridschemas: list[ty.Union[dict, DataGridSchema, BaseModel, ty.Type[BaseModel]]],
) -> tuple[list[xw.worksheet.Worksheet], list[XlTableWriter]]:
    return zip(
        *[
            write_sheet(workbook, data, schema)
            for data, schema in zip(datas, gridschemas)
        ]
    )


def wb_from_json(
    data: list[dict],
    schema: dict,
    fpth: ty.Optional[pathlib.Path] = None,
    exclude_metadata: bool = False,
) -> tuple[xw.Workbook, XlTableWriter, xw.worksheet.Worksheet]:
    if fpth is None:
        fpth = pathlib.Path(schema.get("title") + ".xlsx")
    workbook = xw.Workbook(str(fpth))
    worksheet, xl_tbl = write_sheet(
        workbook,
        data=data,
        gridschema=schema,
        exclude_metadata=exclude_metadata,
    )
    return workbook, xl_tbl, worksheet


def wb_from_jsons(
    datas: list[list[dict]],
    schemas: list[ty.Union[dict, DataGridSchema, BaseModel]],
    fpth: pathlib.Path,
) -> tuple[xw.Workbook, XlTableWriter, xw.worksheet.Worksheet]:
    workbook = xw.Workbook(str(fpth))
    # schemas = coerce_lengths(len(datas), schemas)
    worksheets, xl_tbls = write_sheets(workbook, datas, schemas)
    return workbook, worksheets, xl_tbls


def xdg_from_json(
    data: ty.Union[dict[ty.Union[str, int, float], list], list[dict], pd.DataFrame],
    schema: ty.Union[dict, DataGridSchema, BaseModel],
    fpth: ty.Optional[pathlib.Path] = None,
    is_transposed: ty.Optional[bool] = None,
    exclude_metadata: bool = False,
):
    gridschema = coerce_schema(schema)
    if is_transposed is not None:
        gridschema.is_transposed = is_transposed
    if fpth is None:
        fpth = pathlib.Path(gridschema.title + ".xlsx")
    workbook, xl_tbl, worksheet = wb_from_json(
        data,
        gridschema,
        fpth=fpth,
        exclude_metadata=exclude_metadata,
    )
    workbook.close()
    return fpth, xl_tbl


def xdg_from_jsons(
    datas: list[list[dict]],
    gridschemas: list[ty.Union[dict, DataGridSchema, BaseModel]],
    fpth: pathlib.Path = None,
):
    if fpth is None:
        fpth = pathlib.Path("output" + ".xlsx")
    workbook, worksheets, xl_tbls = wb_from_jsons(datas, gridschemas, fpth)
    workbook.close()
    return fpth


def wb_from_dataframe(
    dataframe: pd.DataFrame, fpth: pathlib.Path, schema: ty.Optional[dict] = None
) -> tuple[xw.Workbook, XlTableWriter, xw.worksheet.Worksheet]:
    schema = (lambda s, df: build_table_schema(df) if s is None else s)(
        schema, dataframe
    )
    if "title" not in schema.keys():
        schema["title"] = fpth.stem
    data = dataframe.reset_index().to_dict(orient="records")
    return wb_from_json(data, schema, fpth)


def coerce_lengths(length: int, *args: ty.Union[ty.Any, list]):
    return [
        (lambda a: length * [a] if not isinstance(a, list) else a)(a) for a in args
    ]  # coerce fixed attributes into list of equal length to lead


def ensure_titles_in_schemas(schemas):
    return [
        (lambda n, s: s | {"title": f"Sheet{n}"} if "title" not in s else s)(n + 1, s)
        for n, s in enumerate(schemas)
    ]


def wb_from_dataframes(
    dataframes: list[pd.DataFrame],
    fpth: pathlib.Path,
    schemas: ty.Optional[ty.Union[dict, list[dict]]] = None,
    titles: list[str] = None,
) -> tuple[xw.Workbook, list[XlTableWriter], list[xw.worksheet.Worksheet]]:
    # resource = Resource(dataframe)
    # data = resource.read_rows()
    # TODO: use frictionless to get data and schema ?
    #       ^ https://github.com/frictionlessdata/frictionless-py/issues/1678

    (schemas,) = coerce_lengths(len(dataframes), schemas)
    schemas = [
        (lambda s, df: build_table_schema(df) if s is None else s)(schema, dataframe)
        for schema, dataframe in zip(schemas, dataframes)
    ]
    if titles is not None:
        schemas = [s | {"title": t} for s, t in zip(schemas, titles)]
    else:
        schemas = ensure_titles_in_schemas(schemas)

    datas = [
        dataframe.reset_index().fillna("").to_dict(orient="records")
        for dataframe in dataframes
    ]
    workbook = xw.Workbook(str(fpth))
    worksheets, xl_tbls = write_sheets(workbook, datas, schemas)

    return workbook, worksheets, xl_tbls


def xdg_from_dataframe(
    dataframe: pd.DataFrame, fpth: pathlib.Path, schema=None
) -> pathlib.Path:
    workbook, xl_tbl, worksheet = wb_from_dataframe(dataframe, fpth, schema)
    workbook.close()
    return fpth


def xdg_from_dataframes(
    dataframes: list[pd.DataFrame],
    fpth: pathlib.Path,
    schemas: ty.Optional[ty.Union[dict, list[dict]]] = None,
    titles: list[str] = None,
) -> pathlib.Path:
    workbook, xl_tbls, worksheets = wb_from_dataframes(
        dataframes, fpth, schemas, titles
    )
    workbook.close()
    return fpth


def get_data_and_dgschema(
    pyd_obj: ty.Type[BaseModel],
) -> tuple[dict[str, list], dict]:
    schema = pyd_obj.model_json_schema(mode="serialization")
    data = pyd_obj.model_dump(mode="json")
    return data, schema


def xdg_from_pydantic_object(
    pydantic_object: ty.Type[BaseModel],
    fpth: pathlib.Path = None,
    is_transposed: ty.Optional[bool] = None,
    exclude_metadata: bool = False,
) -> pathlib.Path:
    data, schema = get_data_and_dgschema(pydantic_object)
    return xdg_from_json(
        data,
        schema,
        fpth=fpth,
        is_transposed=is_transposed,
        exclude_metadata=exclude_metadata,
    )


def xdg_from_pydantic_objects(
    pydantic_objects: list[ty.Type[BaseModel]], fpth: pathlib.Path
) -> pathlib.Path:
    datas, schemas = zip(
        *[
            get_data_and_dgschema(pydantic_object)
            for pydantic_object in pydantic_objects
        ]
    )
    return xdg_from_jsons(datas, schemas, fpth=fpth)



# generated by datamodel-codegen:
#   filename:  test.json
#   timestamp: 2024-09-19T14:16:11+00:00

from __future__ import annotations

from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field


class DEnum(Enum):
    yellow = "yellow"
    red = "red"
    violet = "violet"


class Test(BaseModel):
    a_int: Optional[int] = Field(1, title="A Int")
    DEnum: DEnum = Field(..., description="pick colour")



import pathlib

FDIR = pathlib.Path(__file__).parent / "xl"
PATH_XL, PATH_XL_TRANSPOSED = FDIR / "simple.xlsx", FDIR / "simple-T.xlsx"
PATH_XL_MANY_SHEETS = FDIR / "many-sheets.xlsx"

PATH_XL_FROM_SCHEMA, PATH_XL_FROM_SCHEMA_TRANSPOSED = (
    FDIR / "from-schema.xlsx",
    FDIR / "from-schema-T.xlsx",
)
PATH_XL_FROM_API = FDIR / "from-ds-api.xlsx"
PATH_JSONSCHEMA_RAW = FDIR / "schema-1.json"
PATH_FROM_DF_WITH_CHART = FDIR / "df-x-squared.xlsx"
PATH_FROM_DF_WITH_MANY_CHARTS = FDIR / "many-charts.xlsx"
PATH_FROM_JSON = FDIR / "from-json.xlsx"
PATH_FROM_JSON_EMPTY, PATH_FROM_JSON_EMPTY_TRANSPOSED = (
    FDIR / "from-json.xlsx",
    FDIR / "from-json-T.xlsx",
)
PATH_WRITE_GRID = FDIR / "write-grid.xlsx"



# generated by datamodel-codegen:
#   filename:  example.json
#   timestamp: 2025-10-16T15:55:47+00:00

from __future__ import annotations

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Field, RootModel


class Abbreviation(Enum):
    DB = "DB"


class DistributionBoardItem(BaseModel):
    Abbreviation_1: Optional[Abbreviation] = Field(
        "DB",
        alias="Abbreviation",
        description="equipment type abbreviation, alphabetic characters only, variable length, between 2 and 6 characters",
        title="Abbreviation",
    )
    ClassificationUniclassProductNumber: Optional[str] = Field(
        "Pr_60_70_22_22",
        description="Product Code that matches a code within the Uniclass Pr tables. https://uniclass.thenbs.com/taxon/pr",
        title="Classification Uniclass Product Number",
    )
    ClassificationUniclassSystemNumber: Optional[str] = Field(
        "",
        description="System Code that matches a code within the Uniclass Ss tables. https://uniclass.thenbs.com/taxon/ss",
        title="Classification Uniclass System Number",
    )
    FunctionReference: Optional[str] = Field(
        None,
        description="indicates the functional use of an element",
        title="Function Reference",
    )
    Id: Optional[int] = Field(None, description="", title="Id")
    ManufacturerWebsite: Optional[str] = Field(
        None, description="", title="Manufacturer Website"
    )
    Notes: Optional[str] = Field(
        None, description="free flowing notes section", title="Notes"
    )
    OverallLength: Optional[float] = Field(None, description="", title="Overall Length")
    Symbol: Optional[str] = Field(
        "", description="equipment type symbol. Filename of PNG image", title="Symbol"
    )
    TypeReference: Optional[int] = Field(
        1,
        description="equipment type reference, integer only. Refers to equipment type, there may be multiple instances of the same type",
        title="Type Reference",
    )
    Voltage: Optional[float] = Field(None, description="", title="Voltage")


class DistributionBoard(RootModel):
    root: List[DistributionBoardItem] = Field(..., title="Distribution Board")



from datetime import date, datetime, time, timedelta
from enum import StrEnum

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    NaiveDatetime,
    # NaiveDate,
    RootModel,
    StringConstraints,
    computed_field,
)
from typing_extensions import Annotated


class MyColor(StrEnum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class DataTypes(BaseModel):
    """
    Many data types defined, to be used as a row in excel files
    """

    a_int: int = Field(1, json_schema_extra=dict(section="numeric"))
    a_constrainedint: Annotated[int, Field(ge=0, le=10)] = Field(
        3, json_schema_extra=dict(section="numeric")
    )
    b_float: float = Field(1.5, json_schema_extra=dict(section="numeric"))
    c_str: str = Field("string", json_schema_extra=dict(section="unicode"))
    c_constrainedstr: Annotated[
        str,
        StringConstraints(
            max_length=10,
        ),
    ] = Field("string", json_schema_extra=dict(section="unicode"))
    d_enum: MyColor = Field(json_schema_extra=dict(section="unicode"))
    e_bool: bool = Field(True, json_schema_extra=dict(section="boolean"))
    f_date: date = Field(date.today(), json_schema_extra=dict(section="datetime"))
    g_datetime: NaiveDatetime = Field(
        datetime.now(), json_schema_extra=dict(section="datetime")
    )
    h_time: time = Field(
        datetime.now().time(), json_schema_extra=dict(section="datetime")
    )
    i_duration: timedelta = Field(
        timedelta(days=0, hours=2, minutes=33, seconds=3),
        json_schema_extra=dict(section="datetime"),
    )

    @computed_field(
        description="calc value",
        json_schema_extra=dict(formula="a_int * b_float", section="numeric"),
    )
    def b_calcfloat(self) -> float:
        return self.a_int * self.b_float


class DataTypesBasicFields(BaseModel):
    a_int: int = Field(1, json_schema_extra=dict(section="numeric"))
    a_constrainedint: Annotated[int, Field(ge=0, le=10)] = Field(
        3, json_schema_extra=dict(section="numeric")
    )
    b_float: float = Field(1.5, json_schema_extra=dict(section="numeric"))
    c_str: str = Field("string", json_schema_extra=dict(section="unicode"))
    c_constrainedstr: Annotated[
        str,
        StringConstraints(
            max_length=10,
        ),
    ] = Field("string", json_schema_extra=dict(section="unicode"))
    d_enum: MyColor = Field(json_schema_extra=dict(section="unicode"))


class DataTypesArray(RootModel):
    """
    Array of D Types - Rows of excel file
    """

    model_config = ConfigDict(
        json_schema_extra=dict(
            datagrid_index_name=("section", "title", "name"), is_transposed=False
        )
    )
    root: list[DataTypes]


class DataTypesArrayTransposed(RootModel):
    """
    Array of D Types - Rows of excel file
    """

    model_config = ConfigDict(
        json_schema_extra=dict(
            datagrid_index_name=("section", "title", "name"), is_transposed=True
        )
    )
    root: list[DataTypes]



import pathlib

import pytest

from xlsxdatagrid.read import read_excel
from xlsxdatagrid.xlsxdatagrid import xdg_from_pydantic_object

from .csv_model import DataTypesArrayTransposed


def _as_delimited(text: str, delimiter: str) -> str:
    return text if delimiter == "\t" else text.replace("\t", delimiter)


@pytest.mark.parametrize(
    "is_transposed, exclude_metadata",
    [
        (True, False),
        (True, True),
        (False, False),
        (False, True),
    ],
)
def test_xdg_from_pydantic_object(is_transposed: bool, exclude_metadata: bool):
    data = [
        {
            "a_constrainedint": 3,
            "a_int": 1,
            "b_float": 1.5,
            "c_constrainedstr": "string",
            "c_str": "string",
            "d_enum": "green",
            "e_bool": True,
            "f_date": "2025-10-24",
            "g_datetime": "2025-10-24T11:16:17Z",
            "h_time": "11:16:17Z",
            "i_duration": "PT2H33M3S",
        },
        {
            "a_constrainedint": 3,
            "a_int": 2,
            "b_float": 2.5,
            "c_constrainedstr": "string",
            "c_str": "asdf",
            "d_enum": "green",
            "e_bool": True,
            "f_date": "2025-10-24",
            "g_datetime": "2025-10-24T11:16:17Z",
            "h_time": "11:16:17Z",
            "i_duration": "PT2H33M3S",
        },
        {
            "a_constrainedint": 3,
            "a_int": 3,
            "b_float": 3.5,
            "c_constrainedstr": "string",
            "c_str": "bluey",
            "d_enum": "blue",
            "e_bool": False,
            "f_date": "2025-10-24",
            "g_datetime": "2025-10-24T11:16:17Z",
            "h_time": "11:16:17Z",
            "i_duration": "PT2H33M3S",
        },
    ]

    dest_dir = pathlib.Path("tests/xl/pydantic")
    dest_dir.mkdir(parents=True, exist_ok=True)

    pydantic_obj = DataTypesArrayTransposed(data)
    fpth = (
        dest_dir / f"test_from_pydantic_object-{is_transposed}-{exclude_metadata}.xlsx"
    )

    out_fpth, _ = xdg_from_pydantic_object(
        pydantic_obj,
        fpth=fpth,
        is_transposed=is_transposed,
        exclude_metadata=exclude_metadata,
    )

    assert out_fpth.exists()

    data, errors = read_excel(
        out_fpth,
        is_transposed=is_transposed,
        header_depth=3,
        model=DataTypesArrayTransposed,
    )

    assert not errors
    assert isinstance(data, list)
    assert len(data) == 3



# SPDX-FileCopyrightText: 2024-present jgunstone <j.gunstone@maxfordham.com>
#
# SPDX-License-Identifier: MIT

"""
root package for xlsxdatagrid
expected usage:
```py
import xlsxdatagrid as xdg

xdg.xdg_from_json(...)  # outputs excel file from json data (list of dicts, or dict of columnar arrays)
xdg.xdg_from_dataframe(...)  # outputs excel file from pandas dataframe
etc.
```
"""

from xlsxdatagrid.read import (
    read_csv_string,
    read_excel,
    # read_csv_string_from_metadata,
    read_excel_from_metadata,
)
from xlsxdatagrid.xlsxdatagrid import (
    # wb_from_pydantic_object,
    # wb_from_pydantic_objects,
    wb_from_dataframe,
    wb_from_dataframes,
    wb_from_json,
    wb_from_jsons,
    xdg_from_dataframe,
    xdg_from_dataframes,
    xdg_from_json,
    xdg_from_jsons,
    xdg_from_pydantic_object,
    xdg_from_pydantic_objects,
)

__all__ = [
    "xdg_from_pydantic_object",
    "xdg_from_pydantic_objects",
    "xdg_from_dataframe",
    "xdg_from_dataframes",
    "xdg_from_json",
    "xdg_from_jsons",
    "wb_from_dataframe",
    "wb_from_dataframes",
    "wb_from_json",
    "wb_from_jsons",
    "read_excel_from_metadata",  # TODO: improve functionality here...
    "read_csv_string",
    "csv_from_string",
    "read_excel",
]
from_pydantic_object = xdg_from_pydantic_object  # TODO: delete
from_pydantic_objects = xdg_from_pydantic_objects  # TODO: delete



import csv
import io
import pathlib
import typing as ty
from datetime import date, datetime

from pydantic import BaseModel

from xlsxdatagrid.xlsxdatagrid import (
    DataGridSchema,
    coerce_schema,
    generate_metadata_string,
)


def csv_from_string(
    csv_string: str,
    schema: dict | DataGridSchema | BaseModel | ty.Type[BaseModel],
    include_metadata: bool = True,
    include_header_titles: bool = True,
    string_delimiter: str = "\t",
    fpth: ty.Optional[pathlib.Path] = None,
    is_transposed: bool = False,
):
    """Convert a CSV string to a CSV file with optional schema metadata and header names."""

    # Parse the CSV string
    rows = list(csv.reader(io.StringIO(csv_string), delimiter=string_delimiter))
    if not rows:
        return None

    # Handle header inclusion/exclusion
    if not include_header_titles:
        if is_transposed:
            rows = rows[1:]
        else:
            rows = [row[1:] for row in rows if len(row) > 1]

    # Prepare schema and optional metadata
    gridschema = coerce_schema(schema)
    metadata = generate_metadata_string(gridschema) if include_metadata else None

    # Define output file path
    fpth = fpth or pathlib.Path(f"{gridschema.title}.csv")

    # Write the output file
    with fpth.open("w", newline="", encoding="utf-8") as f:
        writer = csv.writer(f)
        if metadata:
            writer.writerow([metadata])
        writer.writerows(rows)

    return fpth



from __future__ import annotations

from datetime import date, datetime, time, timedelta
from enum import Enum
from typing import List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field, RootModel, StringConstraints
from typing_extensions import Annotated


class MyColor(str, Enum):
    red = "red"
    green = "green"
    blue = "blue"


class DataTypesArrayTransposedItem(BaseModel):
    a_constrainedint: Annotated[int, Field(ge=1, le=10)] = Field(
        3,
        title="A Constrainedint",
        json_schema_extra=dict(
            section="numeric",
        ),
    )

    a_int: Optional[int] = Field(
        1,
        title="A Int",
        json_schema_extra=dict(section="numeric"),
    )

    b_float: Optional[float] = Field(
        1.5,
        title="B Float",
        json_schema_extra=dict(section="numeric"),
    )

    c_constrainedstr: Annotated[str, StringConstraints(min_length=0, max_length=20)] = (
        Field(
            "string",
            title="C Constrainedstr",
            json_schema_extra=dict(
                section="unicode",
            ),
        )
    )

    c_str: Optional[str] = Field(
        "string",
        title="C Str",
        json_schema_extra=dict(section="unicode"),
    )

    d_enum: MyColor = Field(
        "red",
        title="D Enum",
        json_schema_extra=dict(
            section="unicode",
        ),
    )

    e_bool: Optional[bool] = Field(
        True,
        title="E Bool",
        json_schema_extra=dict(section="boolean"),
    )

    f_date: Optional[date] = Field(
        "2024-06-06",
        title="F Date",
        json_schema_extra=dict(section="datetime"),
    )

    g_datetime: Optional[datetime] = Field(
        "2024-06-06T10:42:54.822063",
        title="G Datetime",
        json_schema_extra=dict(section="datetime"),
    )

    h_time: Optional[time] = Field(
        "10:42:54.822257",
        title="H Time",
        json_schema_extra=dict(section="datetime"),
    )

    i_duration: Optional[timedelta] = Field(
        "PT2H33M3S",
        title="I Duration",
        json_schema_extra=dict(section="datetime"),
    )

    model_config = ConfigDict(
        title="Test",
        json_schema_extra=dict(
            required=["d_enum", "b_calcfloat"],
        ),
    )


class DataTypesArrayTransposed(RootModel):
    model_config = ConfigDict(
        title="ExampleArrayTransposed",
        json_schema_extra=dict(
            datagrid_index_name=("section", "title", "name"),
        ),
    )
    root: List[DataTypesArrayTransposedItem]
